@@ -1,3 +1,4 @@
+import time
 import random
 import os
 import signal
@@ -6,7 +7,8 @@ from unittest.mock import patch
 
 from molotov.api import scenario, global_setup
 from molotov.tests.support import (TestLoop, coserver, dedicatedloop, set_args,
-                                   skip_pypy, only_pypy, catch_sleep)
+                                   skip_pypy, only_pypy, catch_sleep,
+                                   dedicatedloop_noclose)
 from molotov.tests.statsd import UDPServer
 from molotov.run import run, main
 from molotov.sharedcounter import SharedCounters
@@ -33,13 +35,8 @@ class TestRunner(TestLoop):
         args.scenario = 'molotov.tests.test_run'
         return args
 
-    @dedicatedloop
+    @dedicatedloop_noclose
     def test_redirect(self):
-        test_loop = asyncio.get_event_loop()
-        test_loop.set_debug(True)
-        test_loop._close = test_loop.close
-        test_loop.close = lambda: None
-
         @scenario(weight=10)
         async def _one(session):
             # redirected
@@ -61,18 +58,15 @@ class TestRunner(TestLoop):
 
         args = self._get_args()
         args.verbose = 2
+        args.max_runs = 2
         with coserver():
             run(args)
 
         self.assertTrue(len(_RES) > 0)
-        test_loop._close()
 
-    @dedicatedloop
+    @dedicatedloop_noclose
     def test_runner(self):
         test_loop = asyncio.get_event_loop()
-        test_loop.set_debug(True)
-        test_loop._close = test_loop.close
-        test_loop.close = lambda: None
 
         @global_setup()
         def something_sync(args):
@@ -103,6 +97,7 @@ class TestRunner(TestLoop):
 
         server_task = asyncio.ensure_future(server.run())
         stop_task = asyncio.ensure_future(stop())
+        args.max_runs = 3
 
         with coserver():
             run(args)
@@ -114,7 +109,6 @@ class TestRunner(TestLoop):
 
         udp = server.flush()
         self.assertTrue(len(udp) > 0)
-        test_loop._close()
 
     @dedicatedloop
     def test_main(self):
@@ -122,44 +116,49 @@ class TestRunner(TestLoop):
             main()
 
     def _test_molotov(self, *args):
+        if '--duration' not in args and '-d' not in args:
+            args = list(args) + ['--duration', '10']
+        rc = 0
         with set_args('molotov', *args) as (stdout, stderr):
             try:
                 main()
-            except SystemExit:
-                pass
-        return stdout.read().strip(), stderr.read().strip()
+            except SystemExit as e:
+                rc = e.code
+        return stdout.read().strip(), stderr.read().strip(), rc
 
     @dedicatedloop
     def test_version(self):
-        stdout, stderr = self._test_molotov('--version')
+        stdout, stderr, rc = self._test_molotov('--version')
         self.assertEqual(stdout, __version__)
 
     @dedicatedloop
     def test_empty_scenario(self):
-        stdout, stderr = self._test_molotov('')
+        stdout, stderr, rc = self._test_molotov('')
         self.assertTrue('Cannot import' in stdout)
 
     @dedicatedloop
     def test_no_scenario(self):
-        stdout, stderr = self._test_molotov()
+        stdout, stderr, rc = self._test_molotov()
         self.assertTrue('Cannot import' in stdout)
 
     @dedicatedloop
     def test_config_no_scenario(self):
-        stdout, stderr = self._test_molotov('-c', '--config', _CONFIG,
-                                            'DONTEXIST')
+        stdout, stderr, rc = self._test_molotov('-c', '--config', _CONFIG,
+                                                'DONTEXIST')
         wanted = "Can't find 'DONTEXIST' in the config"
         self.assertTrue(wanted in stdout)
 
     @dedicatedloop
     def test_config_verbose_quiet(self):
-        stdout, stderr = self._test_molotov('-qv', '--config', _CONFIG)
+        stdout, stderr, rc = self._test_molotov(
+            '-qv', '--config', _CONFIG)
         wanted = "You can't"
         self.assertTrue(wanted in stdout)
 
     @dedicatedloop
     def test_config_no_scenario_found(self):
-        stdout, stderr = self._test_molotov('-c', 'molotov.tests.test_run')
+        stdout, stderr, rc = self._test_molotov(
+            '-c', 'molotov.tests.test_run')
         wanted = "No scenario was found"
         self.assertTrue(wanted in stdout)
 
@@ -170,8 +169,8 @@ class TestRunner(TestLoop):
         async def not_me(session):
             _RES.append(3)
 
-        stdout, stderr = self._test_molotov('-c', '-s', 'blah',
-                                            'molotov.tests.test_run')
+        stdout, stderr, rc = self._test_molotov('-c', '-s', 'blah',
+                                                'molotov.tests.test_run')
         wanted = "Can't find"
         self.assertTrue(wanted in stdout)
 
@@ -186,9 +185,8 @@ class TestRunner(TestLoop):
         async def here_four(session):
             _RES.append(4)
 
-        stdout, stderr = self._test_molotov('-cx', '--max-runs', '2', '-s',
-                                            'me',
-                                            'molotov.tests.test_run')
+        stdout, stderr, rc = self._test_molotov('-cx', '--max-runs', '2', '-s',
+                                                'me', 'molotov.tests.test_run')
         wanted = "SUCCESSES: 2"
         self.assertTrue(wanted in stdout)
         self.assertTrue(_RES, [4, 4])
@@ -200,12 +198,40 @@ class TestRunner(TestLoop):
         async def here_three(session):
             _RES.append(3)
 
-        stdout, stderr = self._test_molotov('-cx', '--max-runs', '2', '-s',
-                                            'here_three',
-                                            'molotov.tests.test_run')
+        stdout, stderr, rc = self._test_molotov('-cx', '--max-runs', '2', '-s',
+                                                'here_three',
+                                                'molotov.tests.test_run')
         wanted = "SUCCESSES: 2"
         self.assertTrue(wanted in stdout)
 
+    @dedicatedloop
+    def test_fail_mode_pass(self):
+
+        @scenario(weight=10)
+        async def here_three(session):
+            _RES.append(3)
+
+        stdout, stderr, rc = self._test_molotov('-cx', '--max-runs', '2',
+                                                '--fail', '1',
+                                                '-s', 'here_three',
+                                                'molotov.tests.test_run')
+        wanted = "SUCCESSES: 2"
+        self.assertTrue(wanted in stdout)
+        self.assertEqual(rc, 0)
+
+    @dedicatedloop
+    def test_fail_mode_fail(self):
+
+        @scenario(weight=10)
+        async def here_three(session):
+            assert False
+
+        stdout, stderr, rc = self._test_molotov('-cx', '--max-runs', '2',
+                                                '--fail', '1', '-s',
+                                                'here_three',
+                                                'molotov.tests.test_run')
+        self.assertEqual(rc, 1)
+
     @only_pypy
     @dedicatedloop
     def test_uvloop_pypy(self):
@@ -222,10 +248,10 @@ class TestRunner(TestLoop):
             return orig_import(name, *args)
 
         with patch('builtins.__import__', side_effect=import_mock):
-            stdout, stderr = self._test_molotov('-cx', '--max-runs', '2',
-                                                '-s',
-                                                'here_three', '--uvloop',
-                                                'molotov.tests.test_run')
+            stdout, stderr, rc = self._test_molotov('-cx', '--max-runs', '2',
+                                                    '-s',
+                                                    'here_three', '--uvloop',
+                                                    'molotov.tests.test_run')
         wanted = "You can't use uvloop"
         self.assertTrue(wanted in stdout)
 
@@ -245,11 +271,11 @@ class TestRunner(TestLoop):
             return orig_import(name, *args)
 
         with patch('builtins.__import__', side_effect=import_mock):
-            stdout, stderr = self._test_molotov('-cx', '--max-runs', '2',
-                                                '--console-update', '0',
-                                                '-s',
-                                                'here_three', '--uvloop',
-                                                'molotov.tests.test_run')
+            stdout, stderr, rc = self._test_molotov('-cx', '--max-runs', '2',
+                                                    '--console-update', '0',
+                                                    '-s',
+                                                    'here_three', '--uvloop',
+                                                    'molotov.tests.test_run')
         wanted = "You need to install uvloop"
         self.assertTrue(wanted in stdout)
 
@@ -261,9 +287,9 @@ class TestRunner(TestLoop):
         async def here_three(session):
             _RES.append(3)
 
-        stdout, stderr = self._test_molotov('-cx', '--max-runs', '2', '-s',
-                                            'here_three', '--uvloop',
-                                            'molotov.tests.test_run')
+        stdout, stderr, rc = self._test_molotov('-cx', '--max-runs', '2', '-s',
+                                                'here_three', '--uvloop',
+                                                'molotov.tests.test_run')
         wanted = "SUCCESSES: 2"
         self.assertTrue(wanted in stdout, stdout)
 
@@ -274,11 +300,11 @@ class TestRunner(TestLoop):
             async def here_three(session):
                 _RES.append(3)
 
-            stdout, stderr = self._test_molotov('--delay', '.6',
-                                                '--console-update', '0',
-                                                '-cx', '--max-runs', '2', '-s',
-                                                'here_three',
-                                                'molotov.tests.test_run')
+            stdout, stderr, rc = self._test_molotov('--delay', '.6',
+                                                    '--console-update', '0',
+                                                    '-cx', '--max-runs', '2',
+                                                    '-s', 'here_three',
+                                                    'molotov.tests.test_run')
             wanted = "SUCCESSES: 2"
             self.assertTrue(wanted in stdout, stdout)
             self.assertEqual(delay, [1, .1, 1, .6, 1, .1, 1, .6, 1])
@@ -290,12 +316,12 @@ class TestRunner(TestLoop):
             async def here_three(session):
                 _RES.append(3)
 
-            stdout, stderr = self._test_molotov('--ramp-up', '10',
-                                                '--workers', '5',
-                                                '--console-update', '0',
-                                                '-cx', '--max-runs', '2', '-s',
-                                                'here_three',
-                                                'molotov.tests.test_run')
+            stdout, stderr, rc = self._test_molotov('--ramp-up', '10',
+                                                    '--workers', '5',
+                                                    '--console-update', '0',
+                                                    '-cx', '--max-runs', '2',
+                                                    '-s', 'here_three',
+                                                    'molotov.tests.test_run')
             # workers should start every 2 seconds since
             # we have 5 workers and a ramp-up
             # the first one starts immediatly, then each worker
@@ -319,14 +345,14 @@ class TestRunner(TestLoop):
                 else:
                     _RES2['succ'] += 1
 
-            stdout, stderr = self._test_molotov('--sizing',
-                                                '--console-update', '0',
-                                                '--sizing-tolerance', '5',
-                                                '-s', 'sizer',
-                                                'molotov.tests.test_run')
+            stdout, stderr, rc = self._test_molotov('--sizing',
+                                                    '--console-update', '0',
+                                                    '--sizing-tolerance', '5',
+                                                    '-s', 'sizer',
+                                                    'molotov.tests.test_run')
 
-            ratio = float(_RES2['fail']) / float(_RES2['succ']) * 100.
-            self.assertTrue(ratio < 15. and ratio >= 5., ratio)
+        ratio = float(_RES2['fail']) / float(_RES2['succ']) * 100.
+        self.assertTrue(ratio < 15. and ratio >= 5., ratio)
 
     @dedicatedloop
     def test_sizing_multiprocess(self):
@@ -352,11 +378,54 @@ class TestRunner(TestLoop):
                     pass
             stdout, stderr = stdout.read().strip(), stderr.read().strip()
 
-            # stdout, stderr = self._test_molotov()
+            # stdout, stderr, rc = self._test_molotov()
             ratio = (float(counters['FAILED'].value) /
                      float(counters['OK'].value) * 100.)
             self.assertTrue(ratio >= 5., ratio)
 
+    @dedicatedloop_noclose
+    def test_statsd_multiprocess(self):
+        test_loop = asyncio.get_event_loop()
+
+        @scenario()
+        async def staty(session):
+            session.statsd.incr('yopla')
+
+        server = UDPServer('127.0.0.1', 9999, loop=test_loop)
+        _stop = asyncio.Future()
+
+        async def stop():
+            await _stop
+            await server.stop()
+
+        server_task = asyncio.ensure_future(server.run())
+        stop_task = asyncio.ensure_future(stop())
+        args = self._get_args()
+        args.verbose = 2
+        args.processes = 2
+        args.max_runs = 5
+        args.duration = 1000
+        args.statsd = True
+        args.statsd_address = 'udp://127.0.0.1:9999'
+        args.single_mode = 'staty'
+        args.scenario = 'molotov.tests.test_run'
+
+        run(args)
+
+        _stop.set_result(True)
+        test_loop.run_until_complete(asyncio.gather(server_task, stop_task))
+        udp = server.flush()
+
+        incrs = 0
+        for line in udp:
+            for el in line.split(b'\n'):
+                if el.strip() == b'':
+                    continue
+                incrs += 1
+
+        # two processes making 5 run each
+        self.assertEqual(incrs, 10)
+
     @dedicatedloop
     def test_timed_sizing(self):
         _RES2['fail'] = 0
@@ -388,14 +457,14 @@ class TestRunner(TestLoop):
                 # forces a switch
                 await asyncio.sleep(0)
 
-            stdout, stderr = self._test_molotov('--sizing',
-                                                '--sizing-tolerance', '5',
-                                                '--console-update', '0',
-                                                '-cs', 'sizer',
-                                                'molotov.tests.test_run')
+            stdout, stderr, rc = self._test_molotov('--sizing',
+                                                    '--sizing-tolerance', '5',
+                                                    '--console-update', '0',
+                                                    '-cs', 'sizer',
+                                                    'molotov.tests.test_run')
 
-            ratio = float(_RES2['fail']) / float(_RES2['succ']) * 100.
-            self.assertTrue(ratio < 20. and ratio > 5., ratio)
+        ratio = float(_RES2['fail']) / float(_RES2['succ']) * 100.
+        self.assertTrue(ratio < 20. and ratio > 5., ratio)
 
     @dedicatedloop
     def test_sizing_multiprocess_interrupted(self):
@@ -415,11 +484,11 @@ class TestRunner(TestLoop):
             os.kill(os.getpid(), signal.SIGINT)
 
         asyncio.ensure_future(_stop())
-        stdout, stderr = self._test_molotov('--sizing', '-p', '3',
-                                            '--sizing-tolerance', '90',
-                                            '--console-update', '0',
-                                            '-s', 'sizer',
-                                            'molotov.tests.test_run')
+        stdout, stderr, rc = self._test_molotov('--sizing', '-p', '3',
+                                                '--sizing-tolerance', '90',
+                                                '--console-update', '0',
+                                                '-s', 'sizer',
+                                                'molotov.tests.test_run')
         self.assertTrue("Sizing was not finished" in stdout)
 
     @dedicatedloop
@@ -432,11 +501,11 @@ class TestRunner(TestLoop):
                 assert resp.status == 200
 
         with coserver():
-            stdout, stderr = self._test_molotov('-cx', '--max-runs', '1',
-                                                '--use-extension=' + ext,
-                                                '-s',
-                                                'simpletest',
-                                                'molotov.tests.test_run')
+            stdout, stderr, rc = self._test_molotov('-cx', '--max-runs', '1',
+                                                    '--use-extension=' + ext,
+                                                    '-s',
+                                                    'simpletest',
+                                                    'molotov.tests.test_run')
         self.assertTrue("=>" in stdout)
         self.assertTrue("<=" in stdout)
 
@@ -450,11 +519,11 @@ class TestRunner(TestLoop):
                 assert resp.status == 200
 
         with coserver():
-            stdout, stderr = self._test_molotov('-cx', '--max-runs', '1',
-                                                '--use-extension=' + ext,
-                                                '-s',
-                                                'simpletest',
-                                                'molotov.tests.test_run')
+            stdout, stderr, rc = self._test_molotov('-cx', '--max-runs', '1',
+                                                    '--use-extension=' + ext,
+                                                    '-s',
+                                                    'simpletest',
+                                                    'molotov.tests.test_run')
         self.assertTrue("Cannot import" in stdout)
 
     @dedicatedloop
@@ -467,11 +536,10 @@ class TestRunner(TestLoop):
                 assert resp.status == 200
 
         with coserver():
-            stdout, stderr = self._test_molotov('-cx', '--max-runs', '1',
-                                                '--use-extension=' + ext,
-                                                '-s',
-                                                'simpletest',
-                                                'molotov.tests.test_run')
+            stdout, stderr, rc = self._test_molotov('-cx', '--max-runs', '1',
+                                                    '--use-extension=' + ext,
+                                                    '-s', 'simpletest',
+                                                    'molotov.tests.test_run')
         self.assertTrue("=>" in stdout)
         self.assertTrue("<=" in stdout)
 
@@ -485,11 +553,10 @@ class TestRunner(TestLoop):
                 assert resp.status == 200
 
         with coserver():
-            stdout, stderr = self._test_molotov('-cx', '--max-runs', '1',
-                                                '--use-extension=' + ext,
-                                                '-s',
-                                                'simpletest',
-                                                'molotov.tests.test_run')
+            stdout, stderr, rc = self._test_molotov('-cx', '--max-runs', '1',
+                                                    '--use-extension=' + ext,
+                                                    '-s', 'simpletest',
+                                                    'molotov.tests.test_run')
         self.assertTrue("Cannot import" in stdout)
 
     @dedicatedloop
@@ -499,9 +566,54 @@ class TestRunner(TestLoop):
         async def here_three(session):
             _RES.append(3)
 
-        stdout, stderr = self._test_molotov('-cx', '--max-runs', '1', '-q',
-                                            '-s',
-                                            'here_three',
-                                            'molotov.tests.test_run')
+        stdout, stderr, rc = self._test_molotov('-cx', '--max-runs', '1', '-q',
+                                                '-s', 'here_three',
+                                                'molotov.tests.test_run')
         self.assertEqual(stdout, '')
         self.assertEqual(stderr, '')
+
+    @dedicatedloop_noclose
+    def test_slow_server_force_shutdown(self):
+        @scenario(weight=10)
+        async def _one(session):
+            async with session.get('http://localhost:8888/slow') as resp:
+                assert resp.status == 200
+                _RES.append(1)
+
+        args = self._get_args()
+        args.duration = 2
+        args.verbose = 2
+        args.max_runs = 1
+        args.force_shutdown = True
+        start = time.time()
+        with coserver():
+            run(args)
+
+        # makes sure the test is stopped even if the server
+        # hangs a socket
+        self.assertTrue(time.time() - start < 4)
+        self.assertTrue(len(_RES) == 0)
+
+    @dedicatedloop_noclose
+    def test_slow_server_graceful(self):
+        @scenario(weight=10)
+        async def _one(session):
+            async with session.get('http://localhost:8888/slow') as resp:
+                assert resp.status == 200
+                _RES.append(1)
+
+        args = self._get_args()
+        args.duration = 2
+        args.verbose = 2
+        args.max_runs = 1
+        # graceful shutdown on the other hand will wait
+        # for the worker completion
+        args.graceful_shutdown = True
+
+        start = time.time()
+        with coserver():
+            run(args)
+
+        # makes sure the test finishes
+        self.assertTrue(time.time() - start > 5)
+        self.assertTrue(len(_RES) == 1)
